import pydrs
import time
import struct

hradcVariant = ['HRADC-FBP','HRADC-FAX-A','HRADC-FAX-B','HRADC-FAX-C','HRADC-FAX-D']

def check_hradc_boarddata(drs,hradc_id):
    print('\n######################################################################')
    print('\nConfigurando placa ' + str(hradc_id+1) + ' em UFM mode...')
    print(drs.ConfigHRADCOpMode(hradc_id,1))
    time.sleep(0.5)

    print('\nExtraindo dados da placa...')
    payload_size   = drs.size_to_hex(1+2) #Payload: ID + hradcID
    hex_hradcID    = drs.double_to_hex(hradc_id)
    send_packet    = drs.ComFunction+payload_size+drs.index_to_hex(31)+hex_hradcID
    send_msg       = drs.checksum(drs.SlaveAdd+send_packet)
    drs.ser.write(send_msg.encode('ISO-8859-1'))
    print(drs.ser.read(6))

    print('\nLendo dados da placa...')
    drs.read_var(drs.index_to_hex(50+hradc_id))
    reply_msg = drs.ser.read(1+1+2+56+1)
    print(reply_msg)
    print(len(reply_msg))
    val = struct.unpack('BBHLLHHHHHHfffffffffB',reply_msg)
    print(val)
    
    try:
        boardData = drs.InitHRADC_BoardData(val[3]+val[4]*pow(2,32),val[5],
                                            val[6],val[7],val[8],val[9],
                                            hradcVariant[val[10]],val[11],
                                            val[12],val[13],val[14],val[15],
                                            val[16],val[17],val[18],val[19])

        print('\n###############################')
        print('### Placa ' + str(hradc_id + 1) + ' ja possui dados ###')
        print('###############################')
    
        print('\nColocando a placa em Sampling mode...')
        print(drs.ConfigHRADCOpMode(hradc_id,0))    
        time.sleep(0.5)
        
        return True
    except:
        print('\n#########################')
        print('### Placa ' + str(hradc_id + 1) + ' sem dados ###')
        print('#########################')
               
        print('\nColocando a placa em Sampling mode...')
        print(drs.ConfigHRADCOpMode(hradc_id,0))
        time.sleep(0.5)
        
        return False

def update_hradc_boarddata(drs,hradc_id,board_data):

    drs.ConfigHRADCOpMode(hradc_id,1)
    time.sleep(0.1)
    drs.EraseHRADC_UFM(hradc_id)
    time.sleep(0.5)
    drs.ResetHRADCBoards(1)
    time.sleep(0.5)
    drs.ResetHRADCBoards(0)
    time.sleep(1.5)
    drs.WriteHRADC_BoardData(hradc_id,board_data)
    
drs = pydrs.SerialDRS()

comport = input('\nDigite a porta serial utilizada: ')

drs.Connect(comport,115200)

n_hradc = int(input('\nDigite o numero de placas HRADC deste bastidor: '))

print('Configurando numero de placas HRADC ...')
drs.Config_nHRADC(n_hradc)

for hradc_id in range(n_hradc):
    if not check_hradc_boarddata(drs,hradc_id):
    
        board_data = drs.InitHRADC_BoardData(serial = 1234567890, day = 27, mon = 9,
                            year = 2019, hour = 4, minutes = 20,
                            variant = 'HRADC-FBP', rburden = 20, calibtemp = 40,
                            vin_gain = 1, vin_offset = 0, iin_gain = 1,
                            iin_offset = 0, vref_p = 5, vref_n = -5, gnd = 0)
    
        update_hradc_boarddata(drs,hradc_id,board_data)
        


"""""" """""" """""" """""" """""" """""" """""" """""" """""" """""" """""" """''
======================================================================
                    Listas de Entidades BSMP
        A posição da entidade na lista corresponde ao seu ID BSMP
======================================================================
""" """""" """""" """""" """""" """""" """""" """""" """""" """""" """""" """""" ""

# FAP
list_fap_soft_interlocks = [
    "DCCT 1 Fault",
    "DCCT 2 Fault",
    "DCCT High Difference",
    "Load Feedback 1 Fault",
    "Load Feedback 2 Fault",
    "IGBTs Current High Difference",
]

list_fap_hard_interlocks = [
    "Load Overcurrent",
    "Load Overvoltage",
    "DCLink Overvoltage",
    "DCLink Undervoltage",
    "Welded Contactor Fault",
    "Opened Contactor Fault",
    "IGBT 1 Overcurrent",
    "IGBT 2 Overcurrent",
    "IIB Itlk",
]

list_fap_iib_interlocks = [
    "Input Overvoltage",
    "Output Overvoltage",
    "IGBT 1 Overcurrent",
    "IGBT 2 Overcurrent",
    "IGBT 1 Overtemperature",
    "IGBT 2 Overtemperature",
    "Driver Overvoltage",
    "Driver 1 Overcurrent",
    "Driver 2 Overcurrent",
    "Driver 1 Error",
    "Driver 2 Error",
    "Inductors Overtemperature",
    "Heat-Sink Overtemperature",
    "DCLink Contactor Fault",
    "Contact Sticking of Contactor",
    "External Interlock",
    "Rack Interlock",
    "High Leakage Current",
    "Board IIB Overtemperature",
    "Module Overhumidity",
]

list_fap_iib_alarms = [
    "Input Overvoltage",
    "Output Overvoltage",
    "IGBT 1 Overcurrent",
    "IGBT 2 Overcurrent",
    "IGBT 1 Overtemperature",
    "IGBT 2 Overtemperature",
    "Driver Overvoltage",
    "Driver 1 Overcurrent",
    "Driver 2 Overcurrent",
    "Inductors Overtemperature",
    "Heat-Sink Overtemperature",
    "High Leakage Current",
    "Board IIB Overtemperature",
    "Module Overhumidity",
]

# FAP-4P
list_fap_4p_soft_interlocks = [
    "DCCT 1 Fault",
    "DCCT 2 Fault",
    "DCCT High Difference",
    "Load Feedback 1 Fault",
    "Load Feedback 2 Fault",
    "IGBTs Current High Difference",
]

list_fap_4p_hard_interlocks = [
    "Load Overcurrent",
    "Load Overvoltage",
    "IGBT 1 Mod 1 Overcurrent",
    "IGBT 2 Mod 1 Overcurrent",
    "IGBT 1 Mod 2 Overcurrent",
    "IGBT 2 Mod 2 Overcurrent",
    "IGBT 1 Mod 3 Overcurrent",
    "IGBT 2 Mod 3 Overcurrent",
    "IGBT 1 Mod 4 Overcurrent",
    "IGBT 2 Mod 4 Overcurrent",
    "Welded Contactor Mod 1 Fault",
    "Welded Contactor Mod 2 Fault",
    "Welded Contactor Mod 3 Fault",
    "Welded Contactor Mod 4 Fault",
    "Opened Contactor Mod 1 Fault",
    "Opened Contactor Mod 2 Fault",
    "Opened Contactor Mod 3 Fault",
    "Opened Contactor Mod 4 Fault",
    "DCLink Mod 1 Overvoltage",
    "DCLink Mod 2 Overvoltage",
    "DCLink Mod 3 Overvoltage",
    "DCLink Mod 4 Overvoltage",
    "DCLink Mod 1 Undervoltage",
    "DCLink Mod 2 Undervoltage",
    "DCLink Mod 3 Undervoltage",
    "DCLink Mod 4 Undervoltage",
    "IIB Mod 1 Itlk",
    "IIB Mod 2 Itlk",
    "IIB Mod 3 Itlk",
    "IIB Mod 4 Itlk",
]

list_fap_4p_iib_interlocks = list_fap_iib_interlocks
list_fap_4p_iib_alarms = list_fap_iib_alarms

# FAP-2P2S
list_fap_2p2s_soft_interlocks = [
    "DCCT 1 Fault",
    "DCCT 2 Fault",
    "DCCT High Difference",
    "Load Feedback 1 Fault",
    "Load Feedback 2 Fault",
    "Arms High Difference",
    "IGBTs Current High Difference",
    "Complementary PS Interlock",
]

list_fap_2p2s_hard_interlocks = [
    "Load Overcurrent",
    "IGBT 1 Mod 1 Overcurrent",
    "IGBT 2 Mod 1 Overcurrent",
    "IGBT 1 Mod 2 Overcurrent",
    "IGBT 2 Mod 2 Overcurrent",
    "IGBT 1 Mod 3 Overcurrent",
    "IGBT 2 Mod 3 Overcurrent",
    "IGBT 1 Mod 4 Overcurrent",
    "IGBT 2 Mod 4 Overcurrent",
    "Welded Contactor Mod 1 Fault",
    "Welded Contactor Mod 2 Fault",
    "Welded Contactor Mod 3 Fault",
    "Welded Contactor Mod 4 Fault",
    "Opened Contactor Mod 1 Fault",
    "Opened Contactor Mod 2 Fault",
    "Opened Contactor Mod 3 Fault",
    "Opened Contactor Mod 4 Fault",
    "DCLink Mod 1 Overvoltage",
    "DCLink Mod 2 Overvoltage",
    "DCLink Mod 3 Overvoltage",
    "DCLink Mod 4 Overvoltage",
    "DCLink Mod 1 Undervoltage",
    "DCLink Mod 2 Undervoltage",
    "DCLink Mod 3 Undervoltage",
    "DCLink Mod 4 Undervoltage",
    "IIB Mod 1 Itlk",
    "IIB Mod 2 Itlk",
    "IIB Mod 3 Itlk",
    "IIB Mod 4 Itlk",
    "Arm 1 Overcurrent",
    "Arm 2 Overcurrent",
]

list_fap_2p2s_iib_interlocks = list_fap_iib_interlocks
list_fap_2p2s_iib_alarms = list_fap_iib_alarms

# FAP 225A
list_fap_225A_soft_interlocks = ["IGBTs Current High Difference"]

list_fap_225A_hard_interlocks = [
    "Load Overcurrent",
    "DCLink Contactor Fault",
    "IGBT 1 Overcurrent",
    "IGBT 2 Overcurrent",
]



# import typing
from siriuspy.bsmp import (
    Entity as _Entity,
)
from siriuspy.bsmp import Types as _Types, BSMPType as _BSMPType
from siriuspy.pwrsupply.bsmp.constants import ConstPSBSMP as _c


import typing as _typing


class Parameter:
    def __init__(
        self,
        var_type: _BSMPType,
        unit: str,
        init: bool,
        Op: bool,
        count: int,
    ):
        self.var_type: _BSMPType = var_type
        self.unit: str = unit
        self.init: bool = init
        self.Op: bool = Op
        self.count: int = count


class Parameters(_Entity):
    """Power supply parameters."""

    _parameters: _typing.Dict[int, Parameter] = {
        # ----- class PS -----
        _c.P_PS_NAME: Parameter(
            count=64,
            var_type=_Types.T_FLOAT,
            unit="",
            init=False,
            Op=True,
        ),
        _c.P_PS_MODEL: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        _c.P_PS_NR_PSMODELS: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        # ----- class Communication -----
        _c.P_COMM_CMD_INTERFACE: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=False,
            Op=True,
        ),
        4: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="bps",
            init=True,
            Op=False,
        ),
        5: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=False,
            Op=True,
        ),
        6: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        7: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        8: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        9: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        10: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="%",
            init=True,
            Op=False,
        ),
        # ----- class Control -----
        11: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="Hz",
            init=True,
            Op=False,
        ),
        12: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="Hz",
            init=True,
            Op=False,
        ),
        13: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        14: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="A/V",
            init=False,
            Op=True,
        ),
        15: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="A/V",
            init=False,
            Op=True,
        ),
        16: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="%",
            init=False,
            Op=True,
        ),
        17: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="%",
            init=False,
            Op=True,
        ),
        # ----- class PWM -----
        18: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="Hz",
            init=True,
            Op=False,
        ),
        19: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="ns",
            init=True,
            Op=False,
        ),
        20: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="pu",
            init=False,
            Op=True,
        ),
        21: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="pu",
            init=False,
            Op=True,
        ),
        22: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="pu",
            init=False,
            Op=True,
        ),
        23: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="pu",
            init=False,
            Op=True,
        ),
        24: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="pu",
            init=False,
            Op=True,
        ),
        # ----- class HRADC -----
        25: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        26: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="MHz",
            init=True,
            Op=False,
        ),
        27: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="Hz",
            init=True,
            Op=False,
        ),
        28: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        29: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        30: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        31: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        32: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        # ----- class SigGen -----
        33: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        34: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        35: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="Hz",
            init=False,
            Op=True,
        ),
        36: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="A/V/%",
            init=False,
            Op=True,
        ),
        37: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="A/V/%",
            init=False,
            Op=True,
        ),
        38: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        # ----- class WfmRef -----
        39: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        40: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
        41: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="Hz",
            init=True,
            Op=False,
        ),
        42: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="A/V/%",
            init=False,
            Op=True,
        ),
        43: Parameter(
            count=4,
            var_type=_Types.T_FLOAT,
            unit="A/V/%",
            init=False,
            Op=True,
        ),
        # ----- class Analog Variables -----
        44: Parameter(
            count=64,
            var_type=_Types.T_FLOAT,
            unit="",
            init=False,
            Op=True,
        ),
        45: Parameter(
            count=64,
            var_type=_Types.T_FLOAT,
            unit="",
            init=False,
            Op=True,
        ),
        # ----- class Debounding manager -----
        46: Parameter(
            count=32,
            var_type=_Types.T_FLOAT,
            unit="us",
            init=True,
            Op=False,
        ),
        47: Parameter(
            count=32,
            var_type=_Types.T_FLOAT,
            unit="us",
            init=True,
            Op=False,
        ),
        48: Parameter(
            count=32,
            var_type=_Types.T_FLOAT,
            unit="us",
            init=True,
            Op=False,
        ),
        49: Parameter(
            count=32,
            var_type=_Types.T_FLOAT,
            unit="us",
            init=True,
            Op=False,
        ),
        # ---- Scope -----
        50: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="Hz",
            init=True,
            Op=False,
        ),
        51: Parameter(
            count=1,
            var_type=_Types.T_FLOAT,
            unit="",
            init=True,
            Op=False,
        ),
    }

    def value_to_load(self, eid: int, value):
        """."""
        _parameter = Parameters._parameters[eid]
        size = _parameter.count
        var_types = [_parameter.var_type] * size
        if eid == 0:
            # power supply name
            value = [float(ord(c)) for c in value]
        load = self._conv_value_to_load(var_types, size, value)
        return load

    def load_to_value(self, eid: int, load):
        """."""
        _parameter = Parameters._parameters[eid]
        size = _parameter.count
        var_types = [_parameter.var_type] * size
        value = self._conv_load_to_value(var_types, load)
        if eid == 0:
            # power supply name
            value = [chr(int(v)) for v in value]
        return value

    @property
    def eids(self):
        """Return entities identifications."""
        return tuple(Parameters._parameters.keys())

    def __getitem__(self, key):
        """."""
        return Parameters._parameters[key]



import logging as _logging
import typing as _typing

__author__ = ""
__version__ = "1.0.0"
__date__ = "Fri, 03 Sep 2021 12:15:17 +0000"

def get_logger(
    name=__file__,
    level: int = _logging.INFO,
    handlers: _typing.Optional[_typing.List[_logging.Handler]] = None,
) -> _logging.Logger:
    """Returns a logger object"""

    logger = _logging.getLogger(name)

    if not len(logger.handlers) and not handlers:
        formatter = _logging.Formatter(
            "[%(asctime)s %(levelname)s %(filename)s:%(lineno)s - %(funcName)s] %(message)s"
        )
        logger.setLevel(level)
        console = _logging.StreamHandler()
        console.setFormatter(formatter)
        logger.addHandler(console)
    return logger



ListVar = ['iLoad1','iLoad2','iMod1','iMod2','iMod3','iMod4','vLoad',
           'vDCMod1','vDCMod2','vDCMod3','vDCMod4','vOutMod1','vOutMod2',
           'vOutMod3','vOutMod4','temp1','temp2','temp3','temp4','ps_OnOff',
           'ps_OpMode','ps_Remote','ps_OpenLoop','ps_SoftInterlocks',
           'ps_HardInterlocks','iRef','wfmRef_Gain','wfmRef_Offset','sigGen_Enable','sigGen_Type',
           'sigGen_Ncycles','sigGenPhaseStart','sigGen_PhaseEnd','sigGen_Freq',
           'sigGen_Amplitude','sigGen_Offset','sigGen_Aux','dp_ID','dp_Class','dp_Coeffs','ps_Model',
           'wfmRef_PtrBufferStart','wfmRef_PtrBufferEnd','wfmRef_PtrBufferK','wfmRef_SyncMode']



from time import time
import typing
import serial
import serial.serialutil


from siriuspy.bsmp.serial import IOInterface as _IOInterface


class TCPInterface(_IOInterface):
    pass


class SerialInterface(_IOInterface):
    def __init__(self, path: str, baudrate: int):
        super().__init__()
        self._port: str = path
        self._baudrate: int = baudrate
        self._serial: typing.Optional[serial.serialutil.SerialBase] = None
        self._encoding: str = "utf-8"

        self.open()

    def open(self) -> None:
        """Open the serial connection"""
        if not self._serial:
            self._serial = serial.Serial(port=self._port, baudrate=self._baudrate)

    def close(self) -> None:
        """Close the serial connection"""
        if self._serial:
            self._serial.close()
            self._serial = None

    def UART_read(self) -> typing.List[str]:
        # @todo: Usar a especificação do bsmp para ler o número correto de bytes
        if not self._serial:
            raise Exception("Serial not defined")

        _response = self._serial.read_all()
        _decoded = _response.decode(self._encoding)
        return [s for s in _decoded]

    def UART_write(
        self, stream: typing.List[str], timeout: float
    ) -> typing.Optional[typing.Any]:
        if not self._serial:
            raise Exception("Serial not defined")

        self._serial.write(stream)

    def UART_request(
        self, stream: typing.List[str], timeout: float
    ) -> typing.Optional[typing.List[str]]:
        """Read and Write OP"""
        self.UART_write(stream, timeout=timeout)
        return self.UART_read()



import typing as _typing

from siriuspy.bsmp import (
    Curve as _C,
    Entities as _E,
    Function as _F,
    Variable as _V,
)
from siriuspy.bsmp import Types as _Types
from siriuspy.pwrsupply.bsmp.constants import ConstPSBSMP as _c

from .parameters import Parameters as _Parameters


class EntitiesPS(_E):
    """PS Entities."""

    _ps_variables: _typing.Tuple = (
        # --- common variables
        # fmt: off
        _V(eid=_c.V_PS_STATUS, waccess=False, count=1, var_type=_Types.T_UINT16),
        _V(eid=_c.V_PS_SETPOINT, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=_c.V_PS_REFERENCE, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=_c.V_COUNTER_SET_SLOWREF, waccess=False, count=128, var_type=_Types.T_CHAR),
        _V(eid=4, waccess=False, count=1, var_type=_Types.T_UINT32),
        _V(eid=5, waccess=False, count=1, var_type=_Types.T_UINT32),
        _V(eid=6, waccess=False, count=1, var_type=_Types.T_UINT16),
        _V(eid=7, waccess=False, count=1, var_type=_Types.T_UINT16),
        _V(eid=8, waccess=False, count=1, var_type=_Types.T_UINT16),
        _V(eid=9, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=10, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=11, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=12, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=13, waccess=False, count=4, var_type=_Types.T_FLOAT),
        _V(eid=14, waccess=False, count=1, var_type=_Types.T_UINT16),
        _V(eid=15, waccess=False, count=1, var_type=_Types.T_UINT16),
        _V(eid=16, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=17, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=18, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=19, waccess=False, count=1, var_type=_Types.T_UINT32),
        _V(eid=20, waccess=False, count=1, var_type=_Types.T_UINT32),
        _V(eid=21, waccess=False, count=1, var_type=_Types.T_UINT32),
        _V(eid=22, waccess=False, count=1, var_type=_Types.T_UINT32),
        _V(eid=23, waccess=False, count=1, var_type=_Types.T_UINT32),
        _V(eid=24, waccess=False, count=1, var_type=_Types.T_UINT32),
        _V(eid=25, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=26, waccess=False, count=1, var_type=_Types.T_FLOAT),
        _V(eid=27, waccess=False, count=1, var_type=_Types.T_UINT32),
        # --- undefined variables
        _V(eid=28, waccess=False, count=1, var_type=_Types.T_UINT8),
        _V(eid=29, waccess=False, count=1, var_type=_Types.T_UINT8),
        _V(eid=30, waccess=False, count=1, var_type=_Types.T_UINT8),
    )
    # fmt: on

    _ps_functions: _typing.Tuple = (
        _F(eid=_c.F_TURN_ON, i_type=(), o_type=(_Types.T_UINT8,)),
        _F(eid=_c.F_TURN_OFF, i_type=(), o_type=(_Types.T_UINT8,)),
        _F(eid=_c.F_OPEN_LOOP, i_type=(), o_type=(_Types.T_UINT8,)),
        _F(eid=_c.F_CLOSE_LOOP, i_type=(), o_type=(_Types.T_UINT8,)),
        _F(
            eid=_c.F_SELECT_OP_MODE,
            i_type=(_Types.T_ENUM,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_RESET_INTERLOCKS,
            i_type=(),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SET_COMMAND_INTERFACE,
            i_type=(_Types.T_ENUM,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SET_SERIAL_TERMINATION,
            i_type=(_Types.T_UINT16,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_UNLOCK_UDC,
            i_type=(_Types.T_UINT16,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_LOCK_UDC,
            i_type=(_Types.T_UINT16,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_CFG_SOURCE_SCOPE,
            i_type=(_Types.T_UINT32,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_CFG_FREQ_SCOPE,
            i_type=(_Types.T_FLOAT,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_CFG_DURATION_SCOPE,
            i_type=(_Types.T_FLOAT,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(eid=_c.F_ENABLE_SCOPE, i_type=(), o_type=(_Types.T_UINT8,)),
        _F(
            eid=_c.F_DISABLE_SCOPE,
            i_type=(),
            o_type=(_Types.T_UINT8,),
        ),
        _F(eid=_c.F_SYNC_PULSE, i_type=(), o_type=()),
        _F(
            eid=_c.F_SET_SLOWREF,
            i_type=(_Types.T_FLOAT,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SET_SLOWREF_FBP,
            i_type=(_Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SET_SLOWREF_READBACK_MON,
            i_type=(_Types.T_FLOAT,),
            o_type=(
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
            ),
        ),
        _F(
            eid=_c.F_SET_SLOWREF_FBP_READBACK_MON,
            i_type=(
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
            ),
            o_type=(
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
            ),
        ),
        _F(
            eid=_c.F_SET_SLOWREF_READBACK_REF,
            i_type=(_Types.T_FLOAT,),
            o_type=(
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
            ),
        ),
        _F(
            eid=_c.F_SET_SLOWREF_FBP_READBACK_REF,
            i_type=(
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
            ),
            o_type=(
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
            ),
        ),
        _F(
            eid=_c.F_RESET_COUNTERS,
            i_type=(),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_CFG_WFMREF,
            i_type=(
                _Types.T_UINT16,
                _Types.T_UINT16,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
            ),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SELECT_WFMREF,
            i_type=(_Types.T_UINT16,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_GET_WFMREF_SIZE,
            i_type=(_Types.T_UINT16,),
            o_type=(_Types.T_UINT16,),
        ),
        _F(eid=_c.F_RESET_WFMREF, i_type=(), o_type=(_Types.T_UINT8,)),
        _F(
            eid=_c.F_CFG_SIGGEN,
            i_type=(
                _Types.T_ENUM,
                _Types.T_UINT16,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
            ),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SET_SIGGEN,
            i_type=(_Types.T_FLOAT, _Types.T_FLOAT, _Types.T_FLOAT),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_ENABLE_SIGGEN,
            i_type=(),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_DISABLE_SIGGEN,
            i_type=(),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SET_PARAM,
            i_type=(
                _Types.T_PARAM,
                _Types.T_UINT16,
                _Types.T_FLOAT,
            ),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_GET_PARAM,
            i_type=(
                _Types.T_PARAM,
                _Types.T_UINT16,
            ),
            o_type=(_Types.T_FLOAT,),
        ),
        _F(
            eid=_c.F_SAVE_PARAM_EEPROM,
            i_type=(
                _Types.T_PARAM,
                _Types.T_UINT16,
                _Types.T_UINT16,
            ),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_LOAD_PARAM_EEPROM,
            i_type=(
                _Types.T_PARAM,
                _Types.T_UINT16,
                _Types.T_UINT16,
            ),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SAVE_PARAM_BANK,
            i_type=(_Types.T_UINT16,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_LOAD_PARAM_BANK,
            i_type=(_Types.T_UINT16,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SET_DSP_COEFFS,
            i_type=(
                _Types.T_DSP_CLASS,
                _Types.T_UINT16,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
                _Types.T_FLOAT,
            ),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_GET_DSP_COEFF,
            i_type=(
                _Types.T_DSP_CLASS,
                _Types.T_UINT16,
                _Types.T_UINT16,
            ),
            o_type=(_Types.T_FLOAT,),
        ),
        _F(
            eid=_c.F_SAVE_DSP_COEFFS_EEPROM,
            i_type=(
                _Types.T_DSP_CLASS,
                _Types.T_UINT16,
                _Types.T_UINT16,
            ),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_LOAD_DSP_COEFFS_EEPROM,
            i_type=(
                _Types.T_DSP_CLASS,
                _Types.T_UINT16,
                _Types.T_UINT16,
            ),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_SAVE_DSP_MODULES_EEPROM,
            i_type=(_Types.T_UINT16,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(
            eid=_c.F_LOAD_DSP_MODULES_EEPROM,
            i_type=(_Types.T_UINT16,),
            o_type=(_Types.T_UINT8,),
        ),
        _F(eid=_c.F_RESET_UDC, i_type=(), o_type=()),
    )

    _ps_curves: _typing.Tuple = (
        _C(
            eid=0,
            waccess=True,
            count=256,
            nblocks=16,
            var_type=_Types.T_FLOAT,
        ),
        _C(
            eid=1,
            waccess=True,
            count=256,
            nblocks=16,
            var_type=_Types.T_FLOAT,
        ),
        _C(
            eid=2,
            waccess=False,
            count=256,
            nblocks=16,
            var_type=_Types.T_FLOAT,
        ),
    )

    _ps_parameters = _Parameters()

    def __init__(self):
        """Call super."""
        super().__init__(self._ps_variables, self._ps_curves, self._ps_functions)

    @property
    def parameters(self):
        """Return pwrsupply parameters."""
        return EntitiesPS._ps_parameters
